#include "KickBall.h"
#include "Events/EventRouter.h"
#include "Events/VisionObjectEvent.h"
#include "Shared/WorldState.h"
#include "Motion/HeadPointerMC.h"
#include "Motion/WalkMC.h"
#include "Shared/WMclass.h"
#include "Shared/ProjectInterface.h"

//! Converts degrees to radians
inline double DtoR(double deg) { return (deg/180.0*M_PI); }

const unsigned int VISUAL_TIMER_ID = 1234;
const float FACE_RIGHT = -80.0;
const float FACE_LEFT = 80.0;

void KickBall::DoStart() {

	BehaviorBase::DoStart();

	
	//previous Area... used to track how close the ball is... default to far away
	previous_area = 999;
	ballArea = 0;
	

	//load sounds so that we can play them
	sndman->LoadFile("sniff.wav");

	//prepare motion manager to move head and walk
	headpointer_id = motman->addPersistentMotion(SharedObject<HeadPointerMC>());
	walker_id = motman->addPersistentMotion(SharedObject<WalkMC>());
	leds_mc->cycle(RobotInfo::ModeGreenLEDMask, 1000, 100.0);			//this will cause his back to flash blue
	leds_mc->cycle(RobotInfo::ModeRedLEDMask, 1000, 100.0);
	leds_id = motman->addPersistentMotion(leds_mc);

	//listen for events from vision (ball) and from buttons
	erouter->addListener(this,EventBase::visObjEGID,ProjectInterface::visOrangeBallSID);
	erouter->addListener(this,EventBase::buttonEGID, HeadFrButOffset);
	
	
	erouter->addTimer(this,1000,1000,true);
	erouter->addTimer(this,VISUAL_TIMER_ID,500,true); //create a new timer, ID=1111, every 500ms, repeat=true
	

	HeadPointerMC * headpointer = (HeadPointerMC*)motman->checkoutMotion(headpointer_id);
	headpointer->setJoints(0,0,0);
	motman->checkinMotion(headpointer_id);
	
}

void KickBall::DoStop() {
	
	erouter->removeListener(this);
	erouter->removeTimer(this);
	motman->removeMotion(headpointer_id);
	motman->removeMotion(walker_id);
	motman->removeMotion(leds_id);

	sndman->ReleaseFile("sniff.wav");
	BehaviorBase::DoStop();

}
//###########################################################
//	Process the event
//###########################################################
void KickBall::processEvent(const EventBase& event) {
	

	if(event.getSourceID() == ProjectInterface::visOrangeBallSID)
	{
		//cout << "Ball seen" << endl;
		time_lost = 0;						//rest the lost timer
		
		//get the location of the ball in AIBO's vision (X,Y) coordinates
		double	horiz =	static_cast<const VisionObjectEvent*>(&event)->getCenterX(); //(rotation) and head pan
		double	vert  =	static_cast<const VisionObjectEvent*>(&event)->getCenterY();		//head tilt
		ballArea = static_cast<const VisionObjectEvent*>(&event)->getArea();			//total visible area
		double headPan_offset = state->outputs[HeadOffset+PanOffset];

		float x,y,z;
		leds_mc->flash(RobotInfo::FaceLEDMask, 100);			//make the leds flash to indicate that the ball is visible
		

		//check to see if AIBO is clost to the ball...if he is, then kick it
		if(ballArea >= DESIRED_AREA && abs(headPan_offset + horiz) < DtoR(35) )  //this will also ensure that the ball is (mostly) in front of AIBO
		{
			cout << "Ball is close.  Kick it!" << endl;
			motman->addPrunableMotion((SharedObject<MediumMotionSequenceMC>)("/ms/data/motion/k_head.mot"));

		}



		x = 180.0f*cos(headPan_offset);			//forward movement
		y = 0.0;									//left/right movement (straife)
		z = 1.0f * sin(horiz + headPan_offset);						//adjust the rotation based on the offset
				
		Walk(x,y,z);
		
		CenterHeadOnBall(horiz, vert);


	}
	else if(event.getGeneratorID()==EventBase::timerEGID && event.getSourceID()==VISUAL_TIMER_ID)
	{
		
		float pan;
		
		time_lost++;			//increase the amount of time the ball has been lost

		if(time_lost > 3)		//if lost for more than 3 seconds then pan head
		{
			if(isLeft)			//if we are facing left, then turn head to the right
				pan = FACE_RIGHT;
			else				//otherwise turn the head left
				pan = FACE_LEFT;
			
			HeadPointerMC * headpointer= (HeadPointerMC*)motman->checkoutMotion(headpointer_id);
			headpointer->setJoints(0,DtoR(pan),0);			
			motman->checkinMotion(headpointer_id);
			isLeft = !isLeft;

			time_lost = 0;				
			Walk(0.0, 0.0, 0.0);

		}
	}


}

void KickBall::CenterHeadOnBall(float x, float y)
{

	double tilt=state->outputs[HeadOffset+TiltOffset]-y*M_PI/7;
	double pan=state->outputs[HeadOffset+PanOffset]-x*M_PI/6;
	//set a bound for the tilt and pan angles
	if(tilt<DtoR(-50))
		tilt=DtoR(-50);
	if(tilt>DtoR(0))
		tilt=DtoR(0);
	if(pan>DtoR(80))
		pan=DtoR(80);
	if(pan<DtoR(-80))
		pan=DtoR(-80);
	HeadPointerMC * headpointer= (HeadPointerMC*)motman->checkoutMotion(headpointer_id);
	headpointer->setJoints(tilt,pan,0);
	motman->checkinMotion(headpointer_id);

	return;
}

void KickBall::Walk(float fwd, float side, float rotate)
{
	//this function will adjust the walk
	MMAccessor<WalkMC>(walker_id)->setTargetVelocity(fwd,side,rotate);
}


//######################################
//Sinusoidal Panning function for OPEN-R
//######################################

/*

	static double panPhase      = 0.0;
	static double tiltPhase		= 0.0;
	const  double deltaPhase    = 0.06;   //0.03;  //let's make AIBO search a little faster; appears to be in radians
	const  double panAmplitude  = 70.0;            // degrees that this joint can move from Origin ...according to AIBO specs however, he can move to 93.0
				    
	if (lastRefPan < 0.0)			//lastRefPan is a global variable that holds the last Pan Joint value
		panPhase = lastRefPan / panAmplitude;
	else
		panPhase = M_PI - lastRefPan / panAmplitude;
	
	
	// Set upper bounds for how fast AIBO's head can move
	const double panDeltaLimit  = deltaPhase * panAmplitude; // degrees / 32ms
	const double tiltDeltaLimit = 5.0;                       // degrees / 32ms
	
	double pan  = panAmplitude * sin(panPhase);		//the head will move side to side as the phase changes
	double tilt = 0.0;							    //Return tilt angle to the origin
	panPhase += deltaPhase;
	
	//call function to move head -> it accepts the target pan and tilt angles and attepts to move from the last angles (assuming that it doesn't exceed the DeltaLimit)	
	//it then sets the lastRef Pan and Tilt angles to be the new angles
	MoveHead(lastRefPan, lastRefTilt, pan, tilt, &lastRefPan, &lastRefTilt,	panDeltaLimit, tiltDeltaLimit );
					
*/