#include "WallFollowing.h"
#include "Events/EventRouter.h"
#include "Events/VisionObjectEvent.h"
#include "Shared/WorldState.h"
#include "Motion/HeadPointerMC.h"
#include "Motion/WalkMC.h"
#include "Shared/WMclass.h"
#include "Shared/ProjectInterface.h"
#include <math.h>

enum STATES {WALKING, DECIDING, TURNING};
enum DIRECTION {LEFT, RIGHT, NOT_SET};

const float FACING_LEFT		= 80.0;
const float FACING_FWD		= 0.0;
const float FACING_RIGHT	= -80.0;
const unsigned int PAN_TIMER	= 1234;
const unsigned int DIR_TIMER	= 1235;
const unsigned int TURN_TIMER	= 1236;
const int TOO_CLOSE		= 75;			//75 mm (3in from wall)
const int TOO_FAR		= 100;			//100mm (4in from wall) 
const int OPENING_DIST		= 254;			//254mm (10in from wall) reasonable?

STATES currState = WALKING;
DIRECTION turnDir = NOT_SET;


inline double DtoR(double deg) { return (deg/180.0*M_PI); }
inline double RtoD(double rad) { return (rad*180.0/M_PI); }

void
WallFollowing::DoStart() {
	BehaviorBase::DoStart(); // do this first

		//load sounds so that we can play them
	
	sndman->LoadFile("3yips.wav");

	//prepare motion manager to move head and walk
	headpointer_id = motman->addPersistentMotion(SharedObject<HeadPointerMC>());
	walker_id = motman->addPersistentMotion(SharedObject<WalkMC>());
	leds_mc->cycle(RobotInfo::ModeGreenLEDMask, 1000, 100.0);			//this will cause his back to flash blue
	leds_mc->cycle(RobotInfo::ModeRedLEDMask, 1000, 100.0);
	leds_id = motman->addPersistentMotion(leds_mc);

	//listen for events from vision (ID ball) and from buttons (switch player type/pause)
	erouter->addListener(this,EventBase::visObjEGID, ProjectInterface::visOrangeBallSID);
	erouter->addListener(this,EventBase::buttonEGID, HeadFrButOffset);
	erouter->addListener(this,EventBase::buttonEGID, MiddleBackButOffset);
	erouter->addListener(this,EventBase::buttonEGID, RearBackButOffset);
	erouter->addListener(this,EventBase::buttonEGID, FrontBackButOffset);

	erouter->addListener(this,EventBase::sensorEGID, FarIRFrameOffset);
	erouter->addListener(this,EventBase::sensorEGID, NearIRFrameOffset);
	
	erouter->addTimer(this,1234, 1000,true);				//pan timer #1234 ticks every 1.5 seconds
	erouter->addTimer(this, 1235, 500, true);				//walking update timer

	//start by having AIBO lift his head
	HeadPointerMC * headpointer= (HeadPointerMC*)motman->checkoutMotion(headpointer_id);
	headpointer->setJoints(0,0,0);			
	motman->checkinMotion(headpointer_id);

	SharedObject<MediumMotionSequenceMC>  p("/ms/data/motion/pounce.pos");
	motman->addPrunableMotion(p);
}

void
WallFollowing::DoStop() {
	erouter->removeListener(this);
	erouter->removeTimer(this);
	motman->removeMotion(headpointer_id);
	motman->removeMotion(walker_id);
	motman->removeMotion(leds_id);
	
	BehaviorBase::DoStop(); // do this last
}

void
WallFollowing::processEvent(const EventBase& event) {


	float farDist = state->sensors[FarIRDistOffset];
	float nearDist = state->sensors[NearIRDistOffset];
	double pan = RtoD(state->outputs[HeadOffset+PanOffset]);
	static int timer_counter =0;
	
	if(event.getSourceID() == HeadFrButOffset)
	{
		if(event.getTypeID() == EventBase::activateETID)
		{
			isBtnPressed = true;
		}
		else if(EventBase::deactivateETID == event.getTypeID() && isBtnPressed)
		{
			isBtnPressed = false;
			isStopped = !isStopped;						//toggle if the AIBO should move or stop
		}

	}
	else if(event.getSourceID() == RearBackButOffset)
	{
		
		cout << "Near distance: " << nearDist << endl;
		cout << "Far distance: " << farDist << endl;
	}

	if(inTransition)
	{
		cout << "In Transition Mode" << endl;
		if(event.getGeneratorID()==EventBase::timerEGID && event.getSourceID()==DIR_TIMER)
		{
			counter++;
			if(counter >= 5)			//walk fwd before turning
			{
				GetToSteppin(10.0, 0.0, DtoR(-45.0));		//move a little bit forward to clear the wall and turn right
				currState = TURNING;
				turnDir = RIGHT;
				inTransition = false;
				leds_mc->set(RobotInfo::ModeGreenLEDMask, 0);				//now both lights are stopped
				leds_mc->cycle(RobotInfo::ModeGreenLEDMask, 1000, 100.0);
				leds_mc->cycle(RobotInfo::ModeRedLEDMask, 1000, 100.0);		//flash the red light again
			}
		}

	}
	

	//start of state seperation
	if(!isStopped && !inTransition)
	{
		if(currState == WALKING)
		{

			cout << "Walking" << endl;
			if(event.getGeneratorID()==EventBase::timerEGID)
			{

				if(event.getSourceID()==PAN_TIMER)			//make the head go back and forth
				{
					PanHead();
					return;
				}
				else if(event.getSourceID()==DIR_TIMER)		//if we need to check out current distance & adjust yet
				{
					//cout << "Pan = " << pan << endl;
					if(abs(pan) > 70.0)		//our output is based on the direction that we are facing
					{
						if(nearDist < TOO_CLOSE) //very close to the right wall			
						{
							cout << "Too close to the wall" << endl;
							//adjust the walk by heading left a little
							GetToSteppin(50.0, 0.0, DtoR(10.0));
							
						}
						else if(nearDist >= TOO_FAR)
						{
							if(nearDist >= OPENING_DIST)
							{
								//there appears to be a large gap(opening to right)
								//turn in that direction and go

								inTransition = true;		//signal the start of this transition
								counter =0;					//reset the counter
								leds_mc->set(RobotInfo::ModeRedLEDMask, 0);			//this will leave only the green light flashing
								
						
								GetToSteppin(50.0, 0.0, 0.0);		//move a little bit forward to clear the wall and turn right
								
								return;
								
							}
							//adjust the walk by heading right a little
							cout << "Too far away from wall...adjusting direction" << endl;
							GetToSteppin(50.0, 0.0, DtoR(-10.0));
							
						}
						else
						{
							//looks good..just go fwd
							GetToSteppin(50.0,0.0,0.0);
							cout << "Walking forward" << endl;
						}

					}
					else if(abs(pan) < 10.0)				//only adjust the distance if we are near an absolute right or centered (head) position
					{
						//we are facing fwd
						if(nearDist < TOO_CLOSE)			//near [100mm] the wall in front then stop (for now)
						{
							GetToSteppin(0.0,0.0,0.0);		//stop walking
							cout << "There's a wall directly in front of me" << endl;
							//check to see if there is an exit to the right 
							SetHeadPan(FACING_RIGHT);
							currState = DECIDING;

						}
						else
						{
							//keep going in the direction that we are going (do nothing different)

						}


					} //  ((abs)pan) < 10.0)


				}

			}
				
		} //currState == WALKING
		else if(currState == DECIDING)
		{
			if(abs(pan) > 70.0)		//check to see if head is facing right yet
			{
				if(nearDist >= OPENING_DIST)	//if there is no opening to the right then we are at a corner and need to turn left
				{
					cout <<  "State = deciding...I see an opening to the right" << endl;
					//turn in that direction and go
					
					GetToSteppin(10.0, 0.0, DtoR(-45.0));		//move a little bit foreward to clear the wall and turn right
					currState = TURNING;
					turnDir = RIGHT;
				}
				else	//otherwise, turn left
				{
					cout << "There is a wall in front and to the right.  Turn left" << endl;
					GetToSteppin(10.0, 0.0, DtoR(45.0));
					currState = TURNING;
					turnDir = LEFT;

				}
				timer_counter=0;
				return;
			}

		} //currState == DECIDING
		else if(currState == TURNING)
		{
			if(event.getGeneratorID()==EventBase::timerEGID && event.getSourceID()==DIR_TIMER)
			{
				timer_counter++;
				if(timer_counter >=4)
				{
					cout << "Done turning.  Walk straight again" << endl;
					currState = WALKING;
					turnDir = NOT_SET;
					GetToSteppin(50.0,0.0,0.0);
					timer_counter = 0;
				}
			}

		} //currState = turning
		else //invalid state
		{
			cout << "INVALID STATE!" << endl;
		}
	}

	//walk until there is a wall directly in front of us

/* unit 10^-6 m -> 1/1000 mm 
	Chest distance sensor (100, 900) mm
	Head distance sensor(near) (50, 500) mm
	Head distance sensor(far) (200, 1500) mm  
*/

}

void WallFollowing::GetToSteppin(float fwd, float side, float rotate)
{
	//this function will adjust the walk
	MMAccessor<WalkMC>(walker_id)->setTargetVelocity(fwd,side,rotate);
}

void WallFollowing::PanHead()
{
	static bool isRight = false;				//this variable will be used to maintain which direction to next look to
	
	if(isRight)
	{
		//look fwd
		SetHeadPan(FACING_FWD);
        		
	}
	else
	{
		//look to the left
		SetHeadPan(FACING_RIGHT);

	}
	isRight = !(isRight);

}

void 
WallFollowing::SetHeadPan(float pan)
{
	//function to set the head outputs
	//assumes that the parameters passed are in degrees

	HeadPointerMC * headpointer= (HeadPointerMC*)motman->checkoutMotion(headpointer_id);
	headpointer->setJoints(0.0,DtoR(pan),0.0);			
	motman->checkinMotion(headpointer_id);
}
